/*

relay.cpp

by Jean-Philippe Civade.
Based on Matthew Williams code K8055 for heyu.

*/


#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <iostream>
#include <stdlib.h>
#include <string.h>
using namespace std;

#define MAX_MESSAGE_LENGTH 255
#define MESSAGE_BUFFER_SIZE (MAX_MESSAGE_LENGTH+1)
#define NUM_DIGITAL_OUTPUT_PORTS 8

#define WITHIN(a,b,c) (((a) >= (b)) && ((a) <= (c)))

char responseBuffer[MESSAGE_BUFFER_SIZE];
char commandBuffer[MESSAGE_BUFFER_SIZE];

int dataSocket=0;
int debug=0;

void handleError(const char *message, int returnCode=1) {
	cerr << message << endl;
	if (errno > 0) {
		cerr << strerror(errno) << endl;
	}
	exit (returnCode);
}

int sendCommand(const char *command, char *response, int maxLength) {
	char buffer[MESSAGE_BUFFER_SIZE];
	response[0]=0;

	if (debug >= 1) {
		cout << "*** Sent: " << command << endl;
	}
	strncpy(buffer,command,maxLength);
	strncat(buffer,"\n",1);

	if (send(dataSocket,buffer,strlen(buffer),0) < 0) {
		handleError("Can't send command");
	}

	int bytesReceived=recv(dataSocket,response, maxLength-1,0);
	if (bytesReceived < 0) {
		handleError("Got bad response from server");
	}
	response[bytesReceived]=0;
	if (debug >= 1) {
		cout << "*** Received: " << response;
	}
	if (strncmp(response,"error",5)==0) {
		handleError("Server claims not to understand me");
	}
	if (strncmp(response,"ok",2) != 0) {
		handleError("Something is wrong, server didn't say everything is ok");
	}

	int returnValue=-1;
	char p0[maxLength];
	char p1[maxLength];
	char p2[maxLength];
	char p3[maxLength];
	char p4[maxLength];

	p0[0]=0;
	p1[0]=0;
	p2[0]=0;
	p3[0]=0;
	p4[0]=0;

	sscanf(response,"%s %s %s %s %s", p0, p1, p2, p3, p4);
	returnValue=atoi(p4);
	return returnValue;
}

void getPID() {
	int value=sendCommand("pid",responseBuffer,MAX_MESSAGE_LENGTH);

	cout << "server PID is " << value << endl;
}

void killAll() {
	cout << "killing server and all clients" << endl;
	sendCommand("kill",responseBuffer,MAX_MESSAGE_LENGTH);
}


void readDigitalValue(int portNum) {
	snprintf(commandBuffer,MAX_MESSAGE_LENGTH,"read digital %i",portNum);
	int value=sendCommand(commandBuffer,responseBuffer,MAX_MESSAGE_LENGTH);
	cout << "digital " << portNum << " is " << value << endl;
}

void readStatus() {
	snprintf(commandBuffer,MAX_MESSAGE_LENGTH,"read status");
	int value=sendCommand(commandBuffer,responseBuffer,MAX_MESSAGE_LENGTH);
	cout << "status "  << " is " << value << endl;
}

void writeDigitalValue(int portNum, int value) {
	snprintf(commandBuffer,MAX_MESSAGE_LENGTH,"write digital %i %i",portNum,value);
	sendCommand(commandBuffer,responseBuffer,MAX_MESSAGE_LENGTH);
	cout << "setting digital port " << portNum << " to " << value << endl;
}

int main(int argc, char *argv[]) {
	struct sockaddr_in serverAddress;
	struct hostent *hostInfo;
	bool showHelp=false;
	bool showPID=false;
	bool killServer=false;
	char hostname[100];
	int port=9998;

	strncpy(hostname,"127.0.0.1",sizeof(hostname)-1);

	hostInfo=gethostbyname(hostname);

	if (hostInfo == 0) {
		handleError("Can't resolve hostname");
	}

	dataSocket=socket(AF_INET,SOCK_STREAM,0);

	if (dataSocket < 0) {
		handleError("Can't create socket");
	}

	serverAddress.sin_family=hostInfo->h_addrtype;
	memcpy((char *)&serverAddress.sin_addr.s_addr,
		hostInfo->h_addr_list[0], hostInfo->h_length);
	serverAddress.sin_port=htons(port);

	if (connect(dataSocket,
		(struct sockaddr *) &serverAddress,
		sizeof(serverAddress)) < 0) {
		handleError("Can't connect to server");
	}

	if (argc == 1) {
		showHelp=true;
	} else {
		for (int i=1; i < argc; i++) {
			if (strncmp(argv[i],"help",4)==0) {
				showHelp=true;
				break;
			} 
			if (strncmp(argv[i],"debug",4)==0) {
				debug++;
				cout << "debug level set to " << debug << endl;
				continue;
			} 
			if (strncmp(argv[i],"kill",4)==0) {
				killServer=true;
				continue;
			} 
			if (strncmp(argv[i],"pid",3)==0) {
				showPID=true;
				continue;
			} 
			if (strncmp(argv[i],"host",4)==0) {
				i++;
				if (i==argc) {
					cerr << "host requires a hostname" << endl;
					showHelp=true;
					break;
				}
				strncpy(hostname,argv[i],sizeof(hostname)-1);
				continue;
			}
			if (strncmp(argv[i],"port",4)==0) {
				i++;
				if (i==argc) {
					cerr << "port requires a port number" << endl;
					showHelp=true;
					break;
				}
				port=atoi(argv[i]);
				continue;
			}
			if (strncmp(argv[i],"read",4)==0) {
				if ((i+1)>=argc) { // we need one more parameter if it is "all"
					cerr << "read required at least one additional parameteri\n";
					showHelp=true;
					break;
				}
				if (strncmp(argv[i+1],"status",5)==0) {
					// read status
					readStatus();
					i++;
					continue;
				}
				if ((i+2)>=argc) { // we need two more parameters
					cerr << "this form of read requires two additional parameters" << endl;
					showHelp=true;
					break;
				}
				int portNum=atoi(argv[i+2]);
				if (strncmp(argv[i+1],"digital",7)==0) {
					if (!WITHIN(portNum,1,NUM_DIGITAL_OUTPUT_PORTS)) {
						cerr << "invalid port number for digital read" << endl;
						showHelp=true;
						break;
					}
					readDigitalValue(portNum);
					i+=2;
					continue;
				}
				cerr << "read requires digital or status" << endl;
				showHelp=true;
				break;
			}
			if (strncmp(argv[i],"write",5)==0) {
				if ((i+3)>=argc) { // we need three more parameters
					cerr << "write requires three additional parameters" << endl;
					showHelp=true;
					break;
				}
				int portNum=atoi(argv[i+2]);
				int value=atoi(argv[i+3]);
				if (strncmp(argv[i+1],"digital",7)==0) {
					if (!WITHIN(portNum,1,NUM_DIGITAL_OUTPUT_PORTS)) {
						cerr << "invalid port number for digital write" << endl;
						showHelp=true;
						break;
					}
					if (!WITHIN(value,0,1)) {
						cerr << "invalid value for digital write" <<endl;
						showHelp=true;
						break;
					}
					writeDigitalValue(portNum,value);
					i+=3;
					continue;
				}
				cerr << "write requires 'digital'" << endl;
				showHelp=true;
				break;
			}
			cerr << "Unknown parameter " << argv[i] << endl;
			showHelp=true;
			break;
		}
	}

	if (showHelp) {
		cout << "relay <parameters>" << endl;
		cout << "  help:     show help" << endl;
		cout << "  debug:    show debug info" << endl;
		cout << "  host:     connect to <host>, default is 127.0.0.1" << endl;
		cout << "  port:     connect to <port>, default is 9998" << endl;
		cout << "  pid:      show server PID" << endl;
		cout << "  kill:     kill server and all clients" << endl;
		cout << "  read:     <digital|status> <port number>" << endl;
		cout << "  write:    <digital> <port number> <value>" << endl;
		return(0);
	}


	if (showPID) {
		getPID();
	}
	if (killServer) {
		killAll();
	}
}
