/*

relayd.cpp

by Jean-Philippe Civade.
Based on Matthew Williams code k8055d for heyu.

+------------------------------------------------------------+
| Serial protocol to drive card is :                         |
+------------------------------------------------------------+
| FF 00 xx	Drive all output to xx state                     |
+------------------------------------------------------------+
| FF yy xx	Drive output yy to xx state                      |
+------------------------------------------------------------+
| xx can be 00 (off) or 01 (on)                              |
| yy can be from 01 (output 1, bit 0) to 08 (output 8, bit 7)|
+------------------------------------------------------------+
*/

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <signal.h>
#include <pthread.h>
#include <iostream>
#include <errno.h>
#include <sys/mman.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

using namespace std;

#define MAX_MESSAGE_LENGTH 2048
#define MESSAGE_BUFFER_SIZE (MAX_MESSAGE_LENGTH+1)
#define MAX_WORDS 10
#define DEFAULT_PORT  9998
#define DEFAULT_DEVICE "/dev/ttyUSB0"

#define WITHIN(a,b,c) (((a) >= (b)) && ((a) <= (c)))

// software maintained relay state
int card_status=0;
// file descriptor for serial port
int fd_serial;

bool inBackground=false;
pthread_mutex_t mutex;


void lockMutex(void) {
	pthread_mutex_lock(&mutex);
}

void unlockMutex(void) {
	pthread_mutex_unlock(&mutex);
}

void handleError(const char *message, int returnCode=1) {
	if (inBackground==false) {
		cerr << message << endl;
		if (errno > 0) {
			cerr << strerror(errno) << endl;
		}
	}
	exit (returnCode);
}

void  writeDigital(int portNum, int value) {
	char iobuff[10];
	int returnCode;

	lockMutex();
	strcpy (iobuff,"\xFF\x00\x00\r");
	iobuff[1]=(unsigned char)portNum;
	iobuff[2]=(unsigned char)value;
	returnCode = write(fd_serial, iobuff, 4);
    if (returnCode < 0)
		handleError ("Can't drive relay outputs");
	if (value==1) { 
		// set bit
		card_status |=  1 << (portNum -1);
		}
	else {
		// clear bit
		card_status &= ~(1 << (portNum -1) );
		}
	// set digital output
	unlockMutex();
}

int readStatus () {
	lockMutex();
	return card_status;
	unlockMutex();
}

int readDigital(int portNum) {
	int value=0;
	lockMutex();
	value = ((card_status & (1<<(portNum-1))) )>>(portNum-1);
	unlockMutex();
	return value;
}

int processCommand(char *command, char *response, int maxLength) {
	int maxWord=0;
	char *word[MAX_WORDS];
	bool inWord=false;
	int timeToDie=0;

	response[0]=0;

	maxLength--; // to account for \n at end of response

	snprintf(response,maxLength-1,"error unknown/invalid command \"%s\"",command);

	unsigned int commandLength=strnlen(command,maxLength);
	for (unsigned int i=0; i < commandLength; i++) {
		if (command[i] == ' ') {
			inWord=false;
			command[i] = 0;
			continue;
		}
		if ((command[i] == '\n') || (command[i]=='\r')) {
			command[i] = 0;
			break;
		}
		if (inWord) {
			continue;
		}
		if (maxWord == MAX_WORDS) {
			break;
		}
		word[maxWord++]=&command[i];
		inWord=true;
	}
	// kill daemon
	if (strncmp(word[0],"kill",maxLength)==0) {
		snprintf (response, maxLength-1, "ok confirmation killing server and all clients");
		timeToDie=2;
	// more clean exit.
	} else if (strncmp(word[0],"exit",maxLength)==0) {
		snprintf (response, maxLength-1, "ok confirmation exitting");
		timeToDie=1;
	// ask for PID
	} else if (strncmp(word[0],"pid",maxLength)==0) {
		snprintf (response, maxLength-1, "ok value server pid %i",getppid());
	// write to port. Only digital is supported. PortNum from 1..8 and value from 0..1
	// Since port reading is not hardware supported, maintain a cached value to allow "read digital" or "read status" (all values at a time). 
	} else if (strncmp(word[0],"write",maxLength)==0) {
		if (maxWord >= 4) {
			int portNum=atoi(word[2]);
			int value=atoi(word[3]);
			if (strncmp(word[1],"digital",maxLength)==0) {
				if (WITHIN(portNum,1,8)) {
					if (WITHIN(value,0,1)) {
						writeDigital(portNum,value);
						snprintf (response, maxLength-1, "ok confirmation setting digital port %i to %i",portNum,value);
					}
				}
			}
		}
	// read from port. Only digital and status ar supported.
	} else if (strncmp(word[0],"read",maxLength)==0) {
		if (maxWord >= 2) {
			int portNum=atoi(word[2]);
			if (strncmp(word[1],"digital",maxLength)==0) {
				if (WITHIN(portNum,1,5)) {
					snprintf (response, maxLength-1, "ok value digital %i %i",portNum,readDigital(portNum));
				}

			} else if (strncmp(word[1],"status",maxLength)==0) {
					snprintf(response,maxLength-1,"ok value status %x", readStatus());
			}
		}
	}
	strncat(response,"\n",1);
	return (timeToDie);
}

void handleConnection (int connectSocket) {
	int receivedBytes=0;
	char portionBuffer[MESSAGE_BUFFER_SIZE];
	char messageBuffer[MESSAGE_BUFFER_SIZE];
	char responseBuffer[MESSAGE_BUFFER_SIZE];
	int messageOffset=0;

	messageBuffer[0]=0;

	while ((receivedBytes=recv(connectSocket, portionBuffer, MAX_MESSAGE_LENGTH, 0)) > 0) {
		portionBuffer[receivedBytes]=0;
		strncat(messageBuffer + messageOffset, portionBuffer, MAX_MESSAGE_LENGTH-messageOffset);
		bool foundCR=false;
		for (int i=0; i < receivedBytes+receivedBytes; i++) {
			if (portionBuffer[i]=='\n') {
				foundCR=true;
				break;
			}
		}
		if (!foundCR) {
			messageOffset+=receivedBytes;
		} else {
			char singleMessage[MESSAGE_BUFFER_SIZE];
			int bytesInBuffer=messageOffset+receivedBytes;
			int startOfSingleMessage=0;

			for (int i=0; i < bytesInBuffer; i++) {
				singleMessage[i-startOfSingleMessage]=messageBuffer[i];
				if (singleMessage[i-startOfSingleMessage]=='\n') {
					singleMessage[i-startOfSingleMessage]=0;
					startOfSingleMessage=i+1; // get ready for next message
					if (messageBuffer[i]=='\r') {
						i++;
						startOfSingleMessage++;
					}


					int timeToDie=processCommand(singleMessage, responseBuffer, MAX_MESSAGE_LENGTH);
					send(connectSocket, responseBuffer, strlen(responseBuffer)+1,0);
					if (timeToDie > 0) {
						if (timeToDie == 1) {
							shutdown (connectSocket,SHUT_RDWR);
							return;
						}
						if (timeToDie == 2) { // everyone dies!
						// close port
						kill(0,SIGTERM);
						}
					}
				}
			}
			messageBuffer[0]=0;
			messageOffset=0;
		}
	}
}


int main (int argc, char *argv[]) {
	int listenPort=DEFAULT_PORT;	// Default = 9998
	int listenSocket=0;
	int connectSocket=0;
	struct sockaddr_in clientAddress;
	socklen_t clientAddressLength=0;
	struct sockaddr_in serverAddress;;
	pid_t childPID;
	char device[20];
	// options du port srie
	struct termios options;
	// return value
	int returnCode;
	// setup default port name
	strncpy(device,DEFAULT_DEVICE,sizeof(device));

	if (argc > 1) { // at least 1 parameter
		if ((strncmp(argv[1],"help",4)==0) || (strncmp(argv[1],"--help",6)==0) || (strncmp(argv[1],"-h",2)==0) || (strncmp(argv[1],"-?",2)==0) ) {
			cout << "Usage : relayd [serial port] [tcp port]" << endl;
			cout << "Fork a daemon to drive relay board." << endl;
			cout <<  endl;
			cout << "Arguments :" << endl;
			cout << "    [serial port]     optional, default value = " << DEFAULT_DEVICE << endl;
			cout << "    [tcp port]        optional, default value = " << DEFAULT_PORT << endl;
			exit (0);
		}
		strncpy(device,argv[1],sizeof(device));
	}

	if (argc > 2) { // at least 2 parameters
		listenPort=atoi(argv[2]);
	}

	// open port
	fd_serial = open(device, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd_serial == -1)  {
		handleError ("Can't open serial port");
		}
    else
		fcntl(fd_serial, F_SETFL, 0);	    

     // Get the current options for the port...
    tcgetattr(fd_serial, &options);
     // Set the baud rates to 9600...
    cfsetispeed(&options, B9600);
    cfsetospeed(&options, B9600);
	// N,8,1
	options.c_cflag &= ~PARENB;
	options.c_cflag &= ~CSTOPB;
	options.c_cflag &= ~CSIZE;
	options.c_cflag |= CS8;
	// no handshake
	options.c_iflag &= ~(IXON | IXOFF | IXANY);
	// Raw Input
	options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
	// Raw Output
	options.c_oflag &= ~OPOST;
    // Enable the receiver and set local mode...
    options.c_cflag |= (CLOCAL | CREAD);
    // And finally Set the new options for the port...
    tcsetattr(fd_serial, TCSANOW, &options);

	// Clear ALL outputs
	returnCode = write(fd_serial, "\xFF\x00\x00\r", 4);
    if (returnCode < 0)
		handleError ("Can't reset relay outputs");


	listenSocket=socket(AF_INET, SOCK_STREAM, 0);

	if (listenSocket < 0) {
		handleError("Can't create listen socket");
	}

	int socketOption=1; // 1=yes
	setsockopt(listenSocket, SOL_SOCKET, SO_REUSEADDR, &socketOption, sizeof(socketOption));

	serverAddress.sin_family = AF_INET;
	serverAddress.sin_addr.s_addr = htonl(INADDR_ANY);
	serverAddress.sin_port=htons(listenPort);

	if (bind(listenSocket,
		(struct sockaddr *) &serverAddress, 
		sizeof(serverAddress)) < 0) {
		handleError("Can't bind listen socket");
	}

	pthread_mutex_init(&mutex,0);

	listen (listenSocket, 5);
	clientAddressLength=sizeof(clientAddress);
	childPID=fork();

	if (childPID == -1) {
		handleError("Can't daemonize myself");
	}
	if (childPID != 0) { // I'm the parent
		cout << "Successfully spawned daemon process for device " << device << " on port " << listenPort << " (PID " << childPID << ")" << endl;
		return 0;  // we're done!
	}
	inBackground=true;
	signal(SIGCLD, SIG_IGN); // allow children to die immediately when terminated

	while (1) {
		connectSocket=accept(listenSocket,
			(struct sockaddr *) &clientAddress,
			&clientAddressLength);

		if (connectSocket < 0) {
			handleError("Can't accept connection");
		}

		//cout << "Connection accepted from " << inet_ntoa(clientAddress.sin_addr);
		//cout << ":" << ntohs (clientAddress.sin_port) << endl;

		childPID=fork();
		if (childPID == -1) {
		  handleError("Can't fork child to handle connection");
		}
		if (childPID == 0) { // I'm the child process
			handleConnection(connectSocket);
			return 0;
		} else { // I'm the parent process
			// do nothing special right now
			wait(NULL);
		}
	}
	return 0;
}
